import org.apache.commons.io.IOUtils
import org.apache.log4j.Logger

import com.atlassian.jira.component.ComponentAccessor

def tempoApiToken = 'e4f41d78-f8e1-407a-9fba-ccafd2c83d86'

def tempoDateFromRequestParameter = requestContext?.getRequestParameter('tempoDateFrom')
def tempoDateToRequestParameter = requestContext?.getRequestParameter('tempoDateTo')
def tempoUserRequestParameter = requestContext?.getRequestParameter('tempoUser') // user key in modern Tempo versions (or username in older ones)

def exportWorklogsOfAllTimes = !(tempoDateFromRequestParameter && tempoDateToRequestParameter)
def exportWorklogsOfAllUsers = false // whether to export the worklog of all users or only the one specified by tempoUserRequestParameter

def tempoDateFrom = tempoDateFromRequestParameter ?: '2000-01-01'
def tempoDateTo = tempoDateToRequestParameter ?: '2099-12-31'
def tempoUser = !exportWorklogsOfAllUsers ? tempoUserRequestParameter : null

tempo = new TempoTool(apiToken: tempoApiToken, baseUrl: requestContext.canonicalBaseUrl, dateFrom: tempoDateFrom, dateTo: tempoDateTo, user: tempoUser, exportWorklogsOfAllTimes: exportWorklogsOfAllTimes, issues: issues)

/**
 * Velocity tool to get Tempo worklogs from the Tempo Servlet API.
 *
 * @see https://tempoplugin.jira.com/wiki/display/TEMPO/Tempo+Servlet+Manual
 */
class TempoTool {
	def log = Logger.getLogger(this.getClass())

	def userManager = ComponentAccessor.userManager

	/* Tempo servlet request parameters. */
	def apiToken
	def baseUrl

	def dateFrom
	def dateTo
	/**
	 * If not <code>null</code>, only the worklogs of this user will be returned.
	 * It is the user key in modern Tempo versions (or the username in older ones).
	 */
	def user

	def exportWorklogsOfAllTimes
	def issues

	/**
	 * Whether to preload all the worklogs in the given timeframe with a single Tempo API call.
	 * Preloading performs better for exports with larger number of issues.
	 * If false, worklogs will be downloaded for each issue separately.
	 */
	def preloadWorklogs
	def preloadedWorklogs = []

	def slurper = new XmlSlurper()

	/**
	 * Returns the Tempo managed worklogs for the passed issue as a <code>List</code> of <code>Maps</code>, or the exception object if something went wrong.
	 * @param issueKey is optional. If not defined, all worklogs in the given timeframe are returned.
	 */
	def getWorklogs(def issueKey) {
		preloadWorklogs = (issues.size() > 10) && !exportWorklogsOfAllTimes
		if(preloadWorklogs) {
			if(!preloadedWorklogs) {
				log.debug("Downloading Tempo worklogs in bulk")
				preloadedWorklogs = callTempoApi()
			}
			return preloadedWorklogs.findAll { it.issue_key == issueKey }
		} else {
			log.debug("Downloading Tempo worklogs for ${issueKey}")
			return callTempoApi(issueKey)
		}
	}

	/**
	 * Calls the Tempo API and returns a list of worklog objects.
	 * @see #getWorklogs(java.lang.String)
	 */
	def callTempoApi(def issueKey) {
		InputStream stream
		try {
			// "user" is the user key in modern Tempo versions (or the username in older ones). "user" must be resolved to the username in any case,
			// as the Tempo Servlet API requires the username. Therefore, assuming that a modern Tempo version is being used, "user" is first used to
			// look up by user key, and if that fails, then it must be a username.
			// Potential conflict (extremely low probability): if "user" is the username (because of an older Tempo version being used),
			// but there is another user with the same key, then this latter user will be found. In that case, the fix is to implement bypassing
			// the UserManager call for that given "user" value.
			def username = userManager.getUserByKey(user)?.username ?: user

			// note: you can't filter by username (the Tempo API can filter either by issue key or by username, but not both)
			def url = "${baseUrl}/plugins/servlet/tempo-getWorklog/?dateFrom=${dateFrom}&dateTo=${dateTo}&format=xml&addBillingInfo=true&addUserDetails=true&tempoApiToken=${apiToken}"
			if(issueKey) {
				url += "&issueKey=${issueKey}"
			} else if(username) {
				url += "&userName=${username}"
			}
			log.debug("Tempo URL" + (issueKey ? " for ${issueKey}" : "") + ": ${url}")
			stream = new URL(url).openStream()
			def xml = IOUtils.toString(stream, "UTF-8")
			log.debug("Tempo XML" + (issueKey ? " for ${issueKey}" : "") + ":\n${xml}")

			def result = slurper.parseText(xml)
			def worklogs = result.worklog
			log.debug("${worklogs.size()} Tempo worklogs found" + (issueKey ? " for ${issueKey}" : ""))

			def worklogObjects = []
			result.worklog.each { w ->
				if((issueKey && username) && (username != w.username.text())) { // w.username is an instance of groovy.util.slurpersupport.NodeChildren
					return
				}
				worklogObjects.add([
						"issue_key"         : w.issue_key.text(),
						"username"          : w.username.text().trim(),
						"hours"             : w.hours.toDouble(),
						"work_date"         : Date.parse("yyyy-MM-dd", w.work_date.text()),
						"work_description"  : w.work_description.text(),
						"billing_key"       : w.billing_key,
						"billing_name"      : w.Billing.@name,
						"full_name"         : w.user_details.full_name.text().trim(),
						"activity_name"     : w.activity_name
				])
			}
			log.debug("Worklogs" + (issueKey ? " for ${issueKey}" : "") + ": ${worklogObjects}")
			return worklogObjects
		} catch(Throwable e) {
			log.error('Failed to download Tempo worklogs', e)
			def failureMessage = "Failed to download Tempo worklogs (have you enabled the Tempo API?): ${e.message}"
			return [[ "issue_key": "", "username": "", "hours": "", "work_date": "", "work_description": failureMessage, "billing_key": "", "billing_name": "", "full_name": "", "activity_name": ""]]
		} finally {
			IOUtils.closeQuietly(stream)
		}
	}
}
